/**
 * Parses JsonPath and KeyPath Expression
 */
class JsonExpression {
  String _expression;
  bool _reversed = false;
  JsonExpression(String expression) {
    this._expression = expression;
    if (this._expression[0] == '!') {
      this._reversed = true;
      this._expression = this._expression.substring(1, this._expression.length);
    }
  }
  /**
   * Convert JsonPath to KeyPath
   *  */
  String toPath() {
    String ex = this._expression;
    if (ex[0] == "\$") {
      if (new RegExp(r"(\.){2,}").hasMatch(ex)) {
        Match match = new RegExp(r"(\.){2,}").allMatches(ex).toList()[0];
        var str = "";
        match.group(0).split("").forEach((item) {
          str += ".*\\/";
        });
        ex = ex.replaceAll(match.group(0).toString(), str);
      }
      ex = ex.replaceAll(new RegExp(r"^\$\."), "^\\/");
      ex = ex.replaceAll(new RegExp(r"^\$"), "^\\/");
      ex = ex.replaceAll(".*", "\\/.*");
      ex = ex.replaceAll("[*]", "\\/.*");
      ex = ex.replaceAll(".", "\\/");
      ex = ex.replaceAll("\\/\*", "\\/\.\*");
      //[3]
      Iterable<Match> matches_3 = new RegExp(r"\[([0-9]+)\]").allMatches(ex);
      matches_3.forEach((match) {
        ex = ex.replaceAll(match.group(0), "\\/[" + match.group(1) + "]");
      });
      if (ex.length > 3 &&
          !(new RegExp(r"\/$").hasMatch(ex)) &&
          !(new RegExp(r"\*$").hasMatch(ex))) {
        ex += '\\/\$';
      }
      //fix for \/\/
      ex = ex.replaceAll("\\/\\/\\/", "\\/");
      ex = ex.replaceAll("\\/\\/", "\\/");
      ex = ex.replaceAll("//", "/");
      return ex;
    } else {
      return ex.replaceAll("\//", "\\/");
    }
  }

  bool isContainer() {
    if (new RegExp(r"^\{").hasMatch(this._expression)) {
      return true;
    } else {
      return false;
    }
  }

  bool contains(List<String> keys) {
    Iterable<Match> matches =
        new RegExp(r"\{(.*)\}").allMatches(this._expression);
    List<String> _keys = matches.first.group(1).split(",");
    return _keys.every((String key) {
      bool has = false;
      JsonExpression path = new JsonExpression(key);
      keys.forEach((String compare) {
        if (new RegExp(path.toPath()).hasMatch(compare)) {
          has = true;
        }
      });
      return has;
    });
  }

  String toString() {
    Iterable<Match> matches =
        new RegExp(r"\{(.*)\}").allMatches(this._expression);
    List<String> _keys = matches.first.group(1).split(",");
    return _keys.toString();
  }

  bool evaluate(List<String> keys, String key) {
    if (this.isContainer()) {
      return _reversed ? !this.contains(keys) : this.contains(keys);
    } else {
      String r = this.toPath();
      return _reversed ? !this._evalutePath(r,key) : this._evalutePath(r,key);
    }
  }

  bool _evalutePath(String je1, String je2) {
    String path = je1
        .replaceAll('^', '')
        .replaceAll('\\', '')
        .replaceFirst(new RegExp(r"\$$"), '');
    String path2 = je2
        .replaceAll('^', '')
        .replaceAll('\\', '')
        .replaceAll('\\', '')
        .replaceFirst(new RegExp(r"\$$"), '');
    List<String> journey1 = path.split("/");
    List<String> journey2 = path2.split("/");
    journey1.removeWhere((String item) {
      return item == null || item == '';
    });
    journey2.removeWhere((String item) {
      return item == null || item == '';
    });
    List<String> first, second;
    String firstPath = '';
    if (journey1.length > journey2.length) {
      first = journey2;
      firstPath = path2;
      second = journey1;
    } else {
      first = journey1;
      second = journey2;
      firstPath = path;
    }
    bool match = true;
    int last = 0;
    int len = first.length;
    for (int i = 0; i < len; i++) {
      Iterable<Match> matchesFirst =
          new RegExp(r"^\[([0-9]+)(,|\:)([0-9]+)\](\+|)$").allMatches(first[i]);
      Iterable<Match> matchesSecond =
          new RegExp(r"^\[([0-9]+)(,|\:)([0-9]+)\](\+|)$")
              .allMatches(second[i]);
      Iterable<Match> matchesThird =
          new RegExp(r"^\[(\:)([0-9]+)\]$").allMatches(first[i]);
      Iterable<Match> matchesFourth =
          new RegExp(r"^\[(\:)([0-9]+)\]$").allMatches(second[i]);
      if (matchesFirst.length > 0) {
        try {
          int start = int.parse(matchesFirst.first.group(1));
          int end = int.parse(matchesFirst.first.group(3));
          int operand = int.parse(second[i]);
          if (!(start <= operand && end >= operand)) {
            match = false;
          }
        } catch (e) {
          match = false;
          last = i;
        }
      } else if (matchesSecond.length > 0) {
        try {
          int start = int.parse(matchesSecond.first.group(1));
          int end = int.parse(matchesSecond.first.group(3));
          int operand = int.parse(first[i]);
          if (!(start <= operand && end >= operand)) {
            match = false;
          }
        } catch (e) {
          match = false;
          last = i;
        }
      } else if (matchesThird.length > 0) {
        try {
          int start = 0;
          int end = int.parse(matchesThird.first.group(2));
          int operand = int.parse(second[i]);
          if (!(start <= operand && end >= operand)) {
            match = false;
          }
        } catch (e) {
          match = false;
          last = i;
        }
      } else if (matchesFourth.length > 0) {
        try {
          int start = 0;
          int end = int.parse(matchesFourth.first.group(2));
          int operand = int.parse(first[i]);
          if (!(start <= operand && end >= operand)) {
            match = false;
          }
        } catch (e) {
          match = false;
          last = i;
        }
      } else if (!(new RegExp("^" + first[i] + "\$").hasMatch(second[i]) ||
              new RegExp("^" + second[i] + "\$").hasMatch(first[i])) &&
          first[i] != '\$' &&
          second[i] != '\$') {
        match = false;
        last = i;
        break;
      }
      last = i;
    }
    bool same = false;
    same = match &&
        ((first.length == second.length) ||
            (firstPath.endsWith(".*") &&
                new RegExp(first.getRange(last, first.length).join("/"))
                    .hasMatch(second.getRange(last, second.length).join("/"))));
    if (first.length == 0) {
    same = match &&
        ((first.length == second.length) ||
            (new RegExp('/'+second.getRange(last, second.length).join("/"))
                .hasMatch('/')));
  }
    return same;
  }
}
