import 'dart:async';

import 'jsontoken.dart';
import 'streamer.dart';

class JsonTokenizer {
  Streamer stream;
  JsonToken token;
  int identity;
  int _combine = 0;
  JsonTokenizer(Streamer stream, {int combine = 0}) {
    this.stream = stream;
    this._combine = combine;
    this.identity = JsonToken.identity();
    this.stream.setGroup(identity);
    this.token = new JsonToken(null, JsonToken.NULL,
        position: 0, group: this.identity, combine: this._combine);
  }
  Future<int> tokenize() async {
    Completer<int> completer = new Completer<int>();

    this.stream.next().then((String response) {
      try {
        if (response == null) {
          JsonToken.close(this.identity);
          completer.complete(1);
        } else {
          response = this.tidy(response);
          String regex =
              "((?<=,)|(?=,))|((?<=\:)|(?=\:))|((?<=\[)|(?=\[))|((?<=\])|(?=\]))|((?<=\{)|(?=\}))|((?<=\")|(?=\"))";
          response.split(new RegExp(regex)).forEach((String rawToken) {
            JsonToken other = new JsonToken(
                rawToken, JsonToken.getType(rawToken),
                group: this.identity, combine: this._combine);
            this.token = this.token = (this.token != null
                    ? this.token
                    : JsonToken.queue[this.identity].values.last)
                .append(other);
          });
          if (!completer.isCompleted) {
            new Timer(new Duration(microseconds: 10), () {
              completer.complete(0);
            });
          }
        }
      } catch (e) {
          new Timer(new Duration(microseconds: 10), () {
            completer.completeError(e);
          });
      }
    }).catchError((err) {
      completer.completeError(err);
    });
    return completer.future;
  }

  //remove completed items
  destroy() {
    List deletables = new List();
    List<int> keys = JsonToken.queue[this.identity].keys.toList();
    int len = keys.length;
    for (int m = 0; m < len; m++) {
      int key = keys[m];
      JsonToken item = JsonToken.queue[this.identity][key];
      if ((item.value == null || item.value == ',' || item.value == ':') &&
          item.position != this.token.position) {
        deletables.add(key);
      } else if (item.position != this.token.position) {
        if (item.parent != null) {
          if (item.parent.closed) {
            item.closed = true;
            deletables.add(key);
          } else if (item.closed == true) {
            deletables.add(key);
          }
        } else if (item.closed == true) {
          deletables.add(key);
        }
        if (item.type == JsonToken.KEY) {
          int next = m + 1;
          if (next < len) {
            JsonToken sub = JsonToken.queue[this.identity][keys[next]];
            if (sub.closed == true) {
              deletables.add(key);
            }
          }
        }
      }
    }
    deletables.forEach((v) {
      if (JsonToken.queue[this.identity].containsKey(v)) {
        JsonToken.queue[this.identity].remove(v);
      }
    });
    //delete sendables
  }

  String tidy(String sequence) {
    return sequence
        .replaceAll(new RegExp(r"(\s+|)\{(\s+|)"), '{')
        .replaceAll(new RegExp(r"(\s+|)\:(\s+|)"), ':')
        .replaceAll(new RegExp(r"(\s+|),(\s+|)"), ',')
        .replaceAll(new RegExp(r"(\s+|)\}(\s+|)"), '}')
        .replaceAll(new RegExp(r"(\s+|)\[(\s+|)"), '[')
        .replaceAll(new RegExp(r"(\s+|)\](\s+|)"), ']')
        .replaceAll(new RegExp(r"[\r\n\t]+(.+[\r\n\t]+.+)[\t\r\n]+"), '');
  }
}
