import 'dart:async';
import 'dart:convert' as JSON;
import 'package:heavylist/heavylist.dart';

import 'complexexpression.dart';
import 'jsonexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'streamer.dart';

class Reader {
  JsonTokenizer tokenizer;
  Function _trailing;
  List<MapEntry<ComplexExpression, Function>> _filters;
  Function _done;
  Function _fail;
  Function _progress;
  Duration _delay;
  int _combine = 0;
  Reader(Streamer stream, {int delay = 0, combine = 0}) {
    this._combine = combine;
    this._delay = new Duration(microseconds: delay);
    this._filters = new List<MapEntry<ComplexExpression, Function>>();
    this.tokenizer = new JsonTokenizer(stream, combine: this._combine);

    JsonToken.event(this.tokenizer.identity).listen(
        (List<MapEntry<String, dynamic>> data) {
      try {
        this._parse(data).then((String d) {
          new Timer(_delay, () {
            this._run();
          });
        }).catchError((err) {
          if (this._fail != null) {
            this._fail(err);
          }
        });
      } catch (e) {
        if (this._fail != null) {
          this._fail(e);
        }
      }
    }, onError: (error) {
      if (this._fail != null) {
        this._fail(error);
      }
    }, onDone: () {
      if (this._done != null) {
        this._done();
      }
    });
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }
  _run() {
    this.tokenizer.tokenize().then((item) {}).catchError((err) {
      if (this._fail != null) {
        this._fail(err);
      }
    });
  }

  Future<String> _parse(List<MapEntry<String, dynamic>> parse) async {
    this.tokenizer.destroy();
    Completer<String> completer = new Completer<String>();
    HeavyList<MapEntry<String, dynamic>> list =
        HeavyList<MapEntry<String, dynamic>>(parse);
    list.loop(new Duration(microseconds: 0),
        (List<MapEntry<String, dynamic>> origin) {
      list.empty();
      if (this._progress != null) {
        this._progress(Streamer.progess(this.tokenizer.identity));
      }
      completer.complete("done");
    }, (MapEntry<String, dynamic> item, Function resume) {
      dynamic firstValue = item.value;
      dynamic firstKey = item.key;
      List<String> ikeys = [];
      if (firstValue.runtimeType.toString().contains("HashMap")) {
        ikeys = item.value.keys.toList();
      }
      try {
        if (this._trailing != null) {
          var response = this._trailing(firstValue, firstKey);
          bool isFuture = response.runtimeType.toString().contains('Future');
          if (isFuture) {
            response.then((dynamic v) {
              HeavyList<MapEntry<ComplexExpression, Function>> sublist =
                  HeavyList<MapEntry<ComplexExpression, Function>>(this._filters);
              sublist.loop(new Duration(microseconds: 0),
                  (List<MapEntry<ComplexExpression, Function>> suborigin) {
                resume();
              }, (MapEntry<ComplexExpression, Function> subitem,
                      Function subresume) {
                if (subitem.key.evaluate(ikeys, firstKey)) {
                  var subresponse = subitem.value(firstValue, firstKey);
                  bool subisFuture =
                      subresponse.runtimeType.toString().contains('Future');
                  if (subisFuture) {
                    subresponse.then((dynamic v) {
                      subresume();
                    }).catchError((error) {
                      completer.completeError(error);
                    });
                  } else {
                    subresume();
                  }
                } else {
                  subresume();
                }
              });
            }).catchError((error) {
              completer.completeError(error);
            });
          } else {
            list.clearRegistered();
            resume();
          }
        } else if (this._filters.length > 0) {
          HeavyList<MapEntry<ComplexExpression, Function>> sublist =
              HeavyList<MapEntry<ComplexExpression, Function>>(this._filters);
          sublist.loop(new Duration(microseconds: 0),
              (List<MapEntry<ComplexExpression, Function>> suborigin) {
            resume();
          }, (MapEntry<ComplexExpression, Function> subitem, Function subresume) {
            if (subitem.key.evaluate(ikeys, firstKey)) {
              var subresponse = subitem.value(firstValue, firstKey);
              bool subisFuture =
                  subresponse.runtimeType.toString().contains('Future');
              if (subisFuture) {
                subresponse.then((dynamic v) {
                  subresume();
                }).catchError((error) {
                  completer.completeError(error);
                });
              } else {
                subresume();
              }
            } else {
              subresume();
            }
          });
        } else {
          list.clearRegistered();
          resume();
        }
      } catch (err) {
        completer.completeError(err);
      }
    });
    return completer.future;
  }

  Reader trailing(Future func(dynamic value, String key)) {
    this._trailing = func;
    return this;
  }

  Reader done(Function func) {
    this._done = func;
    return this;
  }

  Reader fail(Function func(dynamic err)) {
    this._fail = func;
    return this;
  }

  Reader progress(Function func(double progress)) {
    this._progress = func;
    return this;
  }

  Reader filter(
      String expression, Future<dynamic> func(dynamic value, String key)) {
    this._filters.add(new MapEntry(new ComplexExpression(expression), func));
    return this;
  }
}
