import 'dart:async';
import 'dart:convert';
import 'dart:io';

/**
 * Identified chunk for a file based on
 * chunksize
 */
class Streamer {
  int _current;
  int _mod;
  //chunk size
  int _size;
  //current chunk size
  int _chunk;
  int _length;
  //max size in kb of each chunk
  int _max;
  int _total;
  int _loops;
  int _eloops;
  int _start = 0;
  int _end = 0;
  bool _ended = false;
  File file;
  static Map<int, double> _progress = new Map<int, double>();
  int group;
  /**
   * @file file to be parsed
   * @start where to start parsing
   * @end where to stop parsing
   */
  Streamer(File file, {start = 0, end = 0, chunksize = 100}) {
    this.file = file;
    this._current = 0;
    this._chunk = start;
    this._start = start;
    this._end = end;
    this._size = 0;
    this._max = chunksize;
    this.group = group;
    this._loops = 0;
    this._eloops = 0;
  }
  setGroup(int group) {
    this.group = group;
    Streamer._progress[this.group] = 0;
  }

  Future<String> next() async {
    Completer<String> completer = new Completer<String>();
    if (this._size > 0) {
      this._serv().then((String value) {
        completer.complete(value);
      }).catchError((error) {
        completer.completeError(error);
      });
    } else {
      try {
        file.length().then((int value) {
          this._total = value;
          this._mod = value % 1000;
          this._length = (value - this._mod) ~/ 1000;
          this._size = this._length ~/ this._max;
          this._loops = this._size + (this._mod > 0 ? 1 : 0);
          //
          int mod = this._end % 1000;
          int length = (this._end - mod) ~/ 1000;
          int size = length ~/ this._max;
          this._eloops = size + (mod > 0 ? 1 : 0);
          this._serv().then((String value) {
            completer.complete(value);
          }).catchError((error) {
            completer.completeError(error);
          });
        }).catchError((error) {
          completer.completeError(error);
        });
      } catch (e) {
        new Timer(new Duration(microseconds: 0), () {
          completer.completeError(e);
        });
      }
    }
    return completer.future;
  }

  Future<String> _serv() async {
    Completer<String> completer = new Completer<String>();
    bool done = false;
    int end = this._chunk + (this._max * 1000);
    if (this._end == 0) {
      if (this._chunk == 0 && end > this._total) {
        end = this._total;
      } else if (end > this._total) {
        end = this._total;
      }
    } else {
      if (this._chunk == 0 && end > this._end) {
        end = this._end;
      } else if (end > this._end) {
        end = this._end;
      }
    }
    if (this._chunk >= end) {
      done = true;
    }
    if (done || _ended) {
      new Timer(new Duration(microseconds: 0), () {
        completer.complete(null);
      });
    } else {
      StringBuffer bf = new StringBuffer();
      Stream<List<int>> inputStream;
      if (end == this._total) {
        Streamer._progress[this.group] = 100;
        inputStream = file.openRead(this._chunk);
        _ended = true;
      } else {
        if (this._end > 0 && end == this._end) {
          Streamer._progress[this.group] = 100;
        } else if (this._end > 0) {
          Streamer._progress[this.group] = (this._current / this._eloops) * 100;
        } else {
          Streamer._progress[this.group] = (this._current / this._loops) * 100;
        }
        inputStream = file.openRead(this._chunk, end);
      }

      inputStream.transform(utf8.decoder) // Decode bytes to UTF-8.
          .listen((String line) {
        bf.write(line);
      }, onDone: () {
        this._chunk = end;
        this._current++;
        completer.complete(bf.toString());
        bf.clear();
      }, onError: (e) {
        completer.completeError(e);
      });
    }
    return completer.future;
  }

  static double progess(group) {
    return Streamer._progress[group];
  }
}
