import 'dart:async';

import 'package:heavylist/heavylist.dart';

import 'complexexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'streamer.dart';
import 'streamitem.dart';

class StreamReader {
  JsonTokenizer tokenizer;
  StreamController<StreamItem> _trailing = new StreamController<StreamItem>();
  List<MapEntry<ComplexExpression, StreamController<StreamItem>>> _filters;
  StreamController<Null> _done = new StreamController<Null>.broadcast();
  StreamController<dynamic> _fail = new StreamController<dynamic>.broadcast();
  StreamController<double> _progress = new StreamController<double>.broadcast();
  Duration _delay;
  int _combine = 0;
  StreamReader(Streamer stream, {int delay = 0, combine = 0}) {
    this._combine = combine;
    this._delay = new Duration(microseconds: delay);
    this._filters =
        new List<MapEntry<ComplexExpression, StreamController<StreamItem>>>();
    this.tokenizer = new JsonTokenizer(stream, combine: this._combine);

    JsonToken.event(this.tokenizer.identity).listen(
        (List<MapEntry<String, dynamic>> data) {
      try {
        this._parse(data).then((String d) {
          new Timer(_delay, () {
            this._run();
          });
        }).catchError((err) {
          if (this._fail.hasListener) {
            this._fail.add(err);
          }
          this._fail.close();
          //
          this._done.close();
          this._progress.close();
          this._trailing.close();
          this._filters.forEach(
              (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
            item.value.close();
          });
        });
      } catch (e) {
        if (this._fail.hasListener) {
          this._fail.add(e);
        }
        this._fail.close();
        //
        this._done.close();
        this._progress.close();
        this._trailing.close();
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
    }, onError: (error) {
      if (this._fail.hasListener) {
        this._fail.add(error);
      }
      this._fail.close();
      //
      this._done.close();
      this._progress.close();
      this._trailing.close();
      this._filters.forEach(
          (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
        item.value.close();
      });
    }, onDone: () {
      if (this._done.hasListener) {
        this._done.add(null);
      }
      this._done.close();
      //
      this._fail.close();
      this._progress.close();
      this._trailing.close();
      this._filters.forEach(
          (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
        item.value.close();
      });
    });
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }
  _run() {
    this.tokenizer.tokenize().then((item) {}).catchError((err) {
      if (!this._fail.isClosed) {
        if (this._fail.hasListener) {
          this._fail.add(err);
        }
        this._fail.close();
        //
        this._done.close();
        this._progress.close();
        this._trailing.close();
        JsonToken.close(this.tokenizer.identity);
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
    });
  }

  Future<String> _parse(List<MapEntry<String, dynamic>> parse) async {
    this.tokenizer.destroy();
    Completer<String> completer = new Completer<String>();
    HeavyList<MapEntry<String, dynamic>> list =
        HeavyList<MapEntry<String, dynamic>>(parse);
    list.loop(new Duration(microseconds: 0),
        (List<MapEntry<String, dynamic>> origin) {
      list.empty();
      double p = Streamer.progess(this.tokenizer.identity);
      if (this._progress.hasListener) {
        this._progress.add(p);
      }
      if (p >= 100) {
        this._progress.close();
        this._trailing.close();
        this._filters.forEach(
            (MapEntry<ComplexExpression, StreamController<StreamItem>> item) {
          item.value.close();
        });
      }
      completer.complete("done");
    }, (MapEntry<String, dynamic> item, Function resume) {
      dynamic firstValue = item.value;
      dynamic firstKey = item.key;
      List<String> ikeys = [];
      if (firstValue.runtimeType.toString().contains("HashMap")) {
        ikeys = item.value.keys.toList();
      }
      try {
        if (this._trailing.hasListener) {
          this._trailing.add(new StreamItem(firstValue, firstKey));
        }
        HeavyList<MapEntry<ComplexExpression, StreamController<StreamItem>>>
            sublist = HeavyList<
                    MapEntry<ComplexExpression, StreamController<StreamItem>>>(
                this._filters);
        sublist.loop(new Duration(microseconds: 0),
            (List<MapEntry<ComplexExpression, StreamController<StreamItem>>>
                suborigin) {
          resume();
        }, (MapEntry<ComplexExpression, StreamController<StreamItem>> subitem,
                Function subresume) {
          if (subitem.key.evaluate(ikeys, firstKey)) {
            if (subitem.value.hasListener) {
              subitem.value.add(new StreamItem(firstValue, firstKey));
            }
            subresume();
          } else {
            subresume();
          }
        });
      } catch (err) {
        completer.completeError(err);
      }
    });
    return completer.future;
  }

  Stream<StreamItem> get trailing {
    return this._trailing.stream;
  }

  Stream<Null> get done {
    return this._done.stream;
  }

  Stream<dynamic> get fail {
    return this._fail.stream;
  }

  Stream<double> get progress {
    return this._progress.stream;
  }

  Stream<StreamItem> filter(String expression) {
    StreamController<StreamItem> si = new StreamController<StreamItem>();
    this._filters.add(new MapEntry(new ComplexExpression(expression), si));
    return si.stream;
  }
}
